///////////////////////////////////////////////////////////////////////////////
// general used code: WiFi, NTP, DST, SUN, OTA from loop() and other actions //
///////////////////////////////////////////////////////////////////////////////

#ifndef _MY_ST_FUNK_H_
#define _MY_ST_FUNC_H_

#include <WiFi.h>
#include <ArduinoOTA.h>
#include <EEPROM.h>
#include "myST_Init.h"
#include <sESPNOW.h>     // www.kendziorra.nl (simpleLib)
#include <sSun.h>        // www.kendziorra.nl (simpleLib)
#include <sMsecTime.h>   // www.kendziorra.nl (simpleLib)

enum {
  OFF,
  DOWN = OFF, 
  ON,
  UP   = ON,      
  NO,
  DIM  = 2
};

char msg[2][now_maxMsgLen + 1];   // 2 msg buffers
uint8_t saveMitute = 99;  // for test on new minute?
boolean newDay = true,    // is set to true on 04:00h and used on (re)start
        isDST = false,    // DST/Summertime flag
        useDST = true,    // using DST/Summertime
        saveIt = false,   // EEPROM save flag
        flagOTA = true,   // OTA on restart flag
        waitForOTA = false;  // OTA 
time_t otaTime,  // OTA update time window
       nowUTC,   // timestamp UTC/GMT
       nowLOC,   // timestamp Local
       dayTime;  // (this day local time timestamp, no date; hour*3600+minute*60)
const time_t dayStart = 4 * 3600;  // a Item new day start at 04:00  
volatile time_t OTA_Detect = 0;  // OTA touch counter

struct sEpoch ep;     // sMsecTime: sec and msec
struct sDateTime dt;  // sMsecTime: als date and time values
struct DSTDAT dstDat;  // calculation data
struct SUN sun;        // result: .rise and .set

IPAddress local_IP,
          gateway_IP,
          subnet_Mask;

void WiFiOff();
boolean WiFiOn(boolean reStart, boolean useIP);
void sendTo(const char *msg);
void ntpTime(boolean reStart, boolean getIP);
void setOTA();
void nowTimes();
uint8_t timesTest(time_t testTime, time_t timeFrom, time_t timeTo, boolean between);
uint8_t parseFldVal(char * mesage);
void insertSpace(uint8_t item); 
boolean mergList(uint8_t item, uint8_t len, uint8_t *flds);
boolean mergAll(uint8_t item);
void readOrgValues();
void writeItems();
void readItems();
void readTest();
void OTA_Interrupt();
uint32_t getMillis(uint32_t _start);

/** WiFi OFF to switch NTP/OTA/ESP-NOW **/
void WiFiOff() {
  if (WiFi.status() == WL_CONNECTED) {
   WiFi.disconnect(true);
   WiFi.mode(WIFI_OFF);
  }
}

/** WiFi ON to switch NTP/ESP-NOW **/
boolean WiFiOn(boolean reStart, boolean useIP) {
  const IPAddress primDNS(8, 8, 8, 8),  // is needed for SNTP time
                   secDNS(8, 8, 4, 4);
  WiFiOff();
  WiFi.mode(WIFI_AP_STA);
  if (useIP) WiFi.config(local_IP, gateway_IP, subnet_Mask, primDNS, secDNS);
  WiFi.begin(mySSID, myPASSWORD);
  WiFi.setAutoReconnect(true);
  WiFi.setHostname(now_SSID);
  for (uint8_t i = 0; i < 5; i++) {
    uint8_t n = 0;
    while (!WiFi.isConnected() && ++n < 25) delay(200);  // 0...5 sec.
    if (WiFi.isConnected()) break;
    if (i == 4) {
      if (reStart) {
        delay(10000);
        ESP.restart();
      }  
      else return false;
    }
    WiFi.begin();
  }
  WiFi.setSleep(false);
  return true;
}


/** OTA firmware upload **/
void OTA_Interrupt() {
  if (!OTA_Detect) OTA_Detect = 1;
}

void setOTA() {
Serial.printf("setHostname: %s \n", now_SSID);  
  ArduinoOTA.setHostname(now_SSID);
  if (WiFiOn(false, true)) {
    ArduinoOTA.begin();
    waitForOTA = true;
    otaTime = now() + 300;  // wait max. 5 minutes for OTA firmware upload
  }
}

uint8_t timesTest(time_t testTime, time_t timeFrom, time_t timeTo, boolean between) {
  // between false: timeFrom return ON, timeTO return OFF, else return NO
  // between true:  timeFrom to timeTo return ON else return OFF
  uint8_t result = NO;
  if (!between) {  // on timefrom ON, on timeTo OFF, else NOT
    if (testTime == timeFrom) result = ON;
    else if (testTime == timeTo) result = OFF;
  }
  else {  // between timeFrom and timeTo ON, else OFF
    if (testTime < dayStart) testTime += sOneday;  // 00:00 to 04:00 time compensation
    if (timeFrom < dayStart) timeFrom += sOneday;
    if (timeTo < dayStart) timeTo += sOneday;
    if (timeTo < timeFrom) timeTo += sOneday;
    if (testTime >= timeFrom && testTime < timeTo) result = ON;
    else result = OFF;
  }
  return result;
}

/** send message to Node or WS **/
void sendTo(const char *msg) {
  char val[3] = {msg[1], msg[2]};
  if (strcmp(now_Address, "000000") == 0) {  // SERIAL TO WS == ONLY NODE "000000"
    if (strcmp(val, "WS") == 0 && now_macID == 0) {
      Serial2.printf("%s\n", msg);  // to WebSocket Client
      return;
    }
  }
  now_sendData(atoi(val), msg);  // send to MAC "00-19" or over MAC "00" and serial to "WS"
  delay(250);
}

/** get the UTC/GMT time from a time Server **/
void ntpTime(boolean reStart, boolean getIP) {  // setup(): get the time. - loop(): sync the time on night
  uint8_t counter = 0;
  WiFiOn(reStart, !getIP);
  if (getIP) {
    local_IP = WiFi.localIP();
    gateway_IP = WiFi.gatewayIP();
    subnet_Mask = WiFi.subnetMask();
    Serial.printf("NTP/OTA IP: %02d.%02d.%02d.%02d\n", local_IP[0], local_IP[1], local_IP[2], local_IP[3]);
  }
  counter = 0;
  while (!sGetNTP(timeZone)) { 
    delay(1000);
    if (counter++ == 5) {  // get time from Node "00000000"
      WiFiOff();
      delay(200);
      sprintf(msg[0], "|000000|%s|GETTIME|.|", now_Address);
      sendTo(msg[0]);
      return;
    } 
  }
  now_espnowOn();
  delay(500);
  WiFiOff();
}

/** calc all te times we need **/
void nowTimes() {
  // nowUTC, nowLOC    (UNIX timestamp, now-UTC/GMT and now-LOCal date and time) 
  // dayTime           (this day local time timestamp, no date; hour*3600+minute*60)
  // sun.rise, sun.set (this day local time sunRise/sunSet timestamp, no date; hour*3600+minute*60)
  // isDST             (is it DST/Summertime?) (if NOT USING_DST then set "isDST" in "myST_Init.h")
  dt = sTimeDate(false);  // false: local time, true GMT/UTC time
  nowLOC = dt.sec;
  nowUTC = dt.sec - dt.offset;
  time_t t = now();
  Serial.printf("t: %d, nowLOC: %d, nowUTC: %d\n", t, nowLOC, nowUTC);
  Serial.printf("%02u-%02u-%04u %02u:%02u:%02u local time  "
                "weekday: %02u, yweek: %u, yday: %03u\n", 
                day(t), month(t), year(t),
                hour(t), minute(t), second(t), 
                weekday(t), yweek(t), yday(t));  
  Serial.printf("%02u-%02u-%04u %02u:%02u:%02u local time  "
                "weekday: %02u, yweek: %u, yday: %03u\n", 
                day(nowLOC), month(nowLOC), year(nowLOC),
                hour(nowLOC), minute(nowLOC), second(nowLOC), 
                weekday(nowLOC), yweek(nowLOC), yday(nowLOC));  
  Serial.printf("%02u-%02u-%04u %02u:%02u:%02u local time  "
                "weekday: %02u, yweek: %u, yday: %03u\n", 
                day(nowUTC), month(nowUTC), year(nowUTC),
                hour(nowUTC), minute(nowUTC), second(nowUTC), 
                weekday(nowUTC), yweek(nowUTC), yday(nowUTC));  
  if (!useDST) nowLOC -= dt.dst;
  dayTime = dt.hour * sOnehour + dt.minute * sOneminute;  // local day timestamp
  if (dayTime == 4 * sOnehour) {  // is't 04:00? if yes: then sync the time with a NTP Server
    now_espnowOff();
    ntpTime(false, false);
    now_espnowOn();
    newDay = true;
  }
  if (newDay) {
    if (useDST) isDST = dt.dst > 0;  // recalc and set DST
    else isDST = false;
    dstDat.d = dt.day;
    dstDat.m = dt.month;
    dstDat.y = dt.year;
    dstDat.zone = dt.zone;
    dstDat.offset = dt.dst;
    dstDat.lat = lat;
    dstDat.lon = lon;
    sun = calcSun(dstDat);  // a new day? get the new local sun.rise, sun.set timestamps
  }
}

/** get the millis after _start with overflow correction **/
uint32_t getMillis(uint32_t _start) {  
  uint32_t _stop = millis();
  if (_stop < _start) return (uint32_t)4294967295 - _start + _stop + 1;
  else return _stop - _start;
}

#endif
