// important: read readme.h
#ifdef ESP32
  #include <WiFi.h>
#else
  #include <ESP8266WiFi.h>
  #define Serial2 Serial
  //##### IF SERIAL UPLOAD THEN DECONNECT ESP8266->GND<-ESP32 #####
  // or use OTA upload 
#endif
#include <ArduinoOTA.h>
#include <WebSocketsServer.h>

/***** EDIT THIS *****/
void wifiOn(const char * ssid = "edit"@,  // universal ESP WiFi (auto re)connection ...
            const char * password = "edit"@);  // ...(un)comment options in function
const uint16_t wsPort = 81;  // Websocket port
const char wsURI[] = "/",
           wsProtocol[] = "myST_WS";
/***** END EDIT THIS *****/

/** IN/OUT buffers **/
const uint8_t queueBuffers = 5,
              maxMsgLen = 249;  // as ESP-NOW nodes
char inBufSer[maxMsgLen + 1],   // Serial iput
     inBufWS[maxMsgLen + 1],    // WebSocket iput
     recvQueueWS[queueBuffers][maxMsgLen + 1],
     recvQueueSer[queueBuffers][maxMsgLen + 1];
uint8_t recvCounterWS = 0,
        recvCounterSer = 0;
WebSocketsServer webSocket = WebSocketsServer(wsPort, wsURI, wsProtocol);

void webSocketEvent(uint8_t num, WStype_t type, uint8_t * payload, size_t length);
void readSerial2();
void sendAction();

void setup() {
  #ifdef ESP32
    Serial2.begin(115200, SERIAL_8N1, 16, 17);  // rx, tx
  #else
    Serial.begin(115200);
  #endif
  while(!Serial2);
  while(!Serial);
  Serial.println("\n...\n");
  wifiOn();
  ArduinoOTA.setHostname("WSmyST");
  ArduinoOTA.begin();
  webSocket.begin();
  webSocket.onEvent(webSocketEvent);
  IPAddress ip[3] = {WiFi.localIP(), WiFi.gatewayIP(), WiFi.subnetMask()};
  Serial.printf("\n  Hostname: %s\n"
                  "  Static localIP: %3d.%3d.%3d.%3d\n"
                  "       gatewayIP: %3d.%3d.%3d.%3d\n"
                  "      subnetMask: %3d.%3d.%3d.%3d\n\n", 
                  ArduinoOTA.getHostname().c_str(), 
                  ip[0][0], ip[0][1], ip[0][2], ip[0][3],
                  ip[1][0], ip[1][1], ip[1][2], ip[1][3],
                  ip[2][0], ip[2][1], ip[2][2], ip[2][3]);
}

void loop() {
  wifiOn();
  ArduinoOTA.handle();
  webSocket.loop();
  if (Serial2.available()) readSerial2();
  /** send data in queue? **/
  if (recvCounterWS || recvCounterSer) sendAction();
  yield();
}

/** WebSocketServer **/
void webSocketEvent(uint8_t num, WStype_t type, uint8_t * payload, size_t length) {
  switch (type) {
    case WStype_DISCONNECTED:
      break;
    case WStype_CONNECTED: {
        webSocket.sendTXT(num, "Connected");
      }
      break;
    case WStype_TEXT: {
        strcpy(inBufWS, (char*)payload);
        while(inBufWS[0] == ' ') {  // trim <-
          for (uint8_t i = 1; i < strlen(inBufWS); i++) {
            inBufWS[i - 1] = inBufWS[i];
          }
          inBufWS[strlen(inBufWS) - 1] = '\0';
        }
        while(inBufWS[strlen(inBufWS) - 1] == ' ') {  // trim ->
          inBufWS[strlen(inBufWS) - 1] = '\0';
        }
        if (inBufWS[0] == '|' && inBufWS[strlen(inBufWS) - 1] == '|') {
          if (recvCounterWS >= queueBuffers) {
            Serial2.printf("##### recvQueueWS full, we remove oldest data from queue... \n");
            for (uint8_t i = (recvCounterWS - 1); i > 1; i--) strcpy(recvQueueWS[i], recvQueueWS[i - 1]);
            recvCounterWS--;
          }
          strcpy(recvQueueWS[recvCounterWS++], inBufWS);
        }
      }
      break;
  }
}

/** softwareserial input from Node0 module **/
void readSerial2() {
  while (Serial2.available()) {
    char c = Serial2.read();
    if (c == '\n') {
      if (inBufSer[0] == '|' && inBufSer[strlen(inBufSer) - 1] == '|') {
        if (recvCounterSer >= queueBuffers) {
          Serial2.printf("##### recvQueueSer full, we remove oldest data from queue... \n");
          for (uint8_t i = (recvCounterSer - 1); i > 1; i--) strcpy(recvQueueSer[i], recvQueueSer[i - 1]);
          recvCounterSer--;
        }
        strcpy(recvQueueSer[recvCounterSer++], inBufSer);
      }
      inBufSer[0] = '\0';
    }
    else {
      if (c < ' ' || c > '|') inBufSer[0] = '\0';  // cut false char
      else strcat(inBufSer, String(c).c_str());
    }
  }
}

/** data to send **/
void sendAction() {
  if (recvCounterSer) {
    webSocket.broadcastTXT(recvQueueSer[0]);  // Peer A00 data to browser/App
    recvCounterSer--;
    for (uint8_t i = recvCounterSer; i > 0; i--) strcpy(recvQueueSer[i - 1], recvQueueSer[i]);
  }
  if (recvCounterWS) {
    Serial2.printf("%s\n", recvQueueWS[0]);  // WebSocket data to Peer A00
    recvCounterWS--;
    for (uint8_t i = recvCounterWS; i > 0; i--) strcpy(recvQueueWS[i - 1], recvQueueWS[i]);
  }
}

void wifiOn(const char * ssid, const char * password) {
  // universal ESP WiFi (auto re)connection (un)comment options as required
  static bool isInit = false;
  static uint8_t attempt = 0;
  if (!isInit) {  // first time init?
    IPAddress local_IP(192, 168, 178, 180),  // if using a static IP edit the values
            gateway_IP(192, 168, 178,   1),
             subnet_IP(255, 255, 255,   0);
    //IPAddress primDNS(8, 8, 8, 8),  // is needed for SNTP time server ...
    //           secDNS(8, 8, 4, 4);  // ... i think all 8 values are OK
    WiFi.mode(WIFI_STA);      // choose a WiFi mode, Station or/and AccessPoint
    //WiFi.mode(WIFI_AP); 
    //WiFi.mode(WIFI_AP_STA); 
    //WiFi.config(local_IP, gateway_IP, subnet_IP);  // if using a static IP
    //WiFi.config(local_IP, gateway_IP, subnet_IP, primDNS, secDNS); // static IP and SNTP time server
    WiFi.setAutoReconnect(true);
    WiFi.begin(ssid, password);
    for (uint8_t i = 0; i < 5; i++) {  // no connection: keep trying
      uint8_t n = 0;
      while (!WiFi.isConnected() && ++n < 25) delay(200);  // 0...5 sec.
      if (WiFi.isConnected()) break;
      if (i == 4) {
        delay(10000);
        ESP.restart();
      }
      WiFi.begin();
    }
    WiFi.setSleep(false);
    if (WiFi.isConnected()) printf("WiFi connected!\n");  // (un)comment as wished
    isInit = true;
  }
  else {
    if (!WiFi.isConnected()) {
      attempt++;
      WiFi.begin();
      uint8_t n = 0;
      while (!WiFi.isConnected() && ++n < 25) delay(200);  // 0...5 sec.
      if (WiFi.isConnected()) {
        attempt = 0;
        //printf("WiFi reconnected!\n");  // (un)comment as wished
      }
      else if (attempt == 5) ESP.restart();  // (un)comment as wished
    }
  }
}
