////////////////////////////////////////////
// specify the funcion Items of this Node //
////////////////////////////////////////////

// ==========================================================//
//  the functions of this Node                               //
//  sun screen DOWN/UP on temp and lux                       //
//  KaKu 433MHz transmit to switch lamp(s)                   //
// ==========================================================//
//  edit the specific function Items code in:                //
//  -----------------------------------------                //
//  "myST_Start.h"   is called from setup()                  //
//  "myST_Cmd.h"     is called on a received command         //
//  "myST_Actiont.h" is, once per minute, called from loop() //
//  if needed: extra onStart/Command/Action functions        //
//     bellow in this header file                            //
// ==========================================================//

#ifndef _MY_ST_NODE_H_
#define _MY_ST_NODE_H_

#include "myST_Gen.h"
#include <sKaKu.h>
#include <sLuxBH1750.h>
// INSTAL the OneWire-master library from https://github.com/PaulStoffregen/OneWire
#include <sTempDS18B20.h>  // use only a DS18B20 and write it in the parasitic power mode

const uint8_t OTA_Pin = 15,  // OTA upload: touch on PIN-15 or use the WS command "|OTA|" ...
              threshold = 40,  // OTA and home touch threshold
              kakuPin[2] = {4, 32},   // 4:433MHz transmitter, 32:LOW=PowerON/HIGH=PowerOFF(used as GND) 
              luxPin[2] = {21, 22},   // sda, scl, BH1750, I2C
              tempPin = 23,           // DS18B20, + 4.7 KOhm, 1-wire
              screenPin[2] = {5, 2};  // 4N35
const float tempCorrect = -0.5;       // compensate the ESP modules warm up, in +/-n.n*C
const boolean fahrenheit = false;     // false:Celsius
uint8_t posLux = 1,   // gemLux position
        posTmp = 1;   // gemTmp position
int8_t  yoyo = 0,     // anti-YoYo screen minutes counters neg:UP, pos.DOWN
        badW = 0;     // bad weather counter
uint16_t gemLux[13];  // 0:gem. lux  used by myST_Action, 1-12:5 sec.values
double gemTmp[5];     // 0:gem. temp used by myST_Action, 1-4:15 sec.values
time_t intervalLux,
       intervalTmp;
        
// a @ sign mean do NOT change this name
// =====================================
enum {  // this Node Items
  nScreen,  // Item 1
  nLampK1,  // Item 2
  nItems,  //@ the number of items in use on this Node
};
enum {  // basic field names
  nName,    //@
  nMode,    //@
  nStatus,  //@
  nPause    //@
};
enum {  // Item 1 special field names
  nTmpDown1 = nPause +1,  // temperature DOWN
  nLuxDown1,              // Lux DOWN
  nTmpUp1,                // temperature UP
  nLuxUp1,                // Lux UP
  nYoYo1,                 // anti-YoYo time in minutes
  nBadLux1,               // bad wether Lux quick UP
  nFields1  //@
};
enum {  // Item 2 spesial field names
  nCode2 = nPause +1,  // KaKu code
  nPeriod2,            // KaKu period
  nOnHH2,              // morning ON time hh:
  nOnMM2,              //                    mm
  nRise2,              // +/- minutes offset on sunrise
  nSet2,               // +/- minutes offset on sunset
  nOffHH2,             // evening/night OFF time hh:
  nOffMM2,             //                           mm
  nFields2  //@
};
const uint8_t nFields = (nFields1 > nFields2 ? nFields1 : nFields2),  // the maximum number of used Node Items fields
              itemFields[nItems] = {nFields1, nFields2},  // the number of fields Items[0], Item[1}, ... , Item[n]
              nMaxChars = (10 + 1);  // max number of chars in the <...Fields[]> and <...Values[]>
const uint16_t maxItemLen = nFields * nMaxChars + nFields + 1,
               eepromAddr[nItems] = {10, (10 + maxItemLen)};  // Item EEPROM start addresses (start after: OTA[0], address[1-6], and useDST[7])
char fldIn[nFields][nMaxChars],  // parse received message to fieldnames an fieldValues 
     valIn[nFields][nMaxChars];  //   ...
const char itemNames[nItems][nFields][nMaxChars] =  {{"name",   "mode", "status", "pause", "tmpDown", "luxDown", "tmpUp", "luxUp",   "yoyo", "badLux"},
                                                     {"name",   "mode", "status", "pause", "code",    "period",  "on-hh", "on-mm",   "rise", "set",    "off-hh", "off-mm"}},
           itemValOrg[nItems][nFields][nMaxChars] = {{"screen", "1",    "0",       "0",    "21.5",    "600",     "21",    "70",      "10",   "35"},
                                                     {"lampK1", "2",    "0",       "0",    "118124",  "322",     "06",    "30",      "15",   "15",     "02",     "15"}};
char itemValues[nItems][nFields][nMaxChars];        // lampK1: mode 1:morning, mode 2:evening, mode 3:morning and evening
char cmdList[nItems][11] = {{"TEMP|LUX"}, {"nothing|."}};  // set additional command list or "nothing", put code in myST_Cmd.h->onCommand() 

/** general used functions **/
void fromSetup();  // wath to do from setup()?
void fromLoop_1();   // wath to do from loop() before onAction()?
void fromLoop_2();   // wath to do from loop() after onAction()?
uint8_t parseFldVal(char * mesage);  // used by onCommand() to parse the input char string
 
///////////////////////////////////////
/** spesial functions for this Thing **/
char inBufSer[now_maxMsgLen + 1];  // Serial iput buffer
void readSerial2();
void screenDownUp(uint8_t item);  // optocoupler bridge, two momentary ON signals 
void kakuOffOn(uint8_t item);  // 433MHz radio
///////////////////////////////////////

/** to do from setup() here **/
void fromSetup() {
  /** init this Node Item(s) pins and ... here **/
  Serial2.begin(115200, SERIAL_8N1, 16, 17);  // rx, tx
  Wire.begin(luxPin[0], luxPin[1]);
  sLuxInit();
//  sInitDS(tempPin);  // tempCorrection default is 0.0
  sInitDS(tempPin, tempCorrect);
  sKaKuInit(kakuPin[0], kakuPin[1]);  // 433MHz transmitter
  gemLux[0] = sGetLux();
  for (uint8_t i = 1; i <= 12; i++) gemLux[i] = gemLux[0];  // first time fill all
  float t = sGetTempDS();
  gemTmp[0] = (fahrenheit ? (t * 9/5 + 32) : t);
  for (uint8_t i = 1; i <= 4; i++) gemTmp[i] = gemTmp[0];   // first time fill all
}
 
/** to do from loop() here **/
void fromLoop_1() {
  // before onAction()
}
 
/** to do from loop() here **/
void fromLoop_2() {
  // after onAction()
  if(Serial2.available()) readSerial2();  // Serial messages from WebSocket browser Client?
  if (now() > intervalLux) {  // get gem. lux and test on bad weather
    if (posLux > 12) posLux = 1;
    gemLux[posLux++] = sGetLux();
    if (atoi(itemValues[nScreen][nStatus])) {  // screen is DOWN?
      if (gemLux[posLux - 1] > atoi(itemValues[nScreen][nBadLux1])) badW = 0;  // reset counter
      else {  // it's to dark
        badW++;
        if (badW == 3) {  // 3x5 sec. bad weather darknis?
          strcpy(itemValues[nScreen][nStatus], "0");  // set value to UP
          screenDownUp(nScreen);
          badW = 0;  // reset counter
          saveIt = true;
        }    
      }
    }
    uint32_t t = 0;
    for (uint8_t i = 1; i <= 12; i++) t += gemLux[i];
    gemLux[0] =  t / 12;  // gem. lux in one minute
    intervalLux = now() + 5;  // 5 sec interval
  }
  if (now() > intervalTmp) {  // get gem. room temperature
    if (posTmp > 4) posTmp = 1;
    double t = sGetTempDS();
    gemTmp[posTmp++] = (fahrenheit ? (t * 9/5 + 32) : t);
    gemTmp[0] = (gemTmp[1] + gemTmp[2] + gemTmp[3] + gemTmp[4]) / 4;
    intervalTmp = now() + 15;  // 15 sec interval
  }  // gemLux[0] and gemTmp[0] are used in "myST_Action.h"
}

/** parse the received message to fieldnames an fieldValues **/ 
uint8_t parseFldVal(char * mesage) {
  uint8_t n = 0;
  char * v;
  /** get the number of fields **/
  v = strchr(mesage, '|');
  while (v != NULL) {
    n++;
    v = strchr(v + 1, '|');
  }
  if (n > 3 && (n % 2 != 0)) { // <= 3:no data, 'n' not neg. wrong format
    n = (n - 1) / 2;;
    uint8_t i = 0;
    /** get the fields **/
    v = strtok (mesage, "|");
    while (v != NULL) {
      strcpy(fldIn[i], v);
      v = strtok (NULL, "|");
      strcpy(valIn[i++], v);
      v = strtok (NULL, "|");
    }
  }
  else n = 0;
  return n;
}

///////////////////////////////////////
// == extra Node function here == //

void screenDownUp(uint8_t item) {  // optocoupler bridge, two momentary ON signals 
  uint8_t status = atoi(itemValues[item][nStatus]);
  pinMode(screenPin[status], OUTPUT);  // some pins are changed by system
  for (uint8_t i = 0; i < 2; i++) { 
    digitalWrite(screenPin[status], HIGH);  // optocoupler ON
    delay(500);
    digitalWrite(screenPin[status], LOW);   // optocoupler OFF
    if (i < 2) delay(500);
  }
  yoyo = 0;
}

// == extra KaKu function == //
void kakuOffOn(uint8_t item) {  // 433MHz radio
  uint8_t status = atoi(itemValues[item][nStatus]);
  uint16_t period = atoi(itemValues[item][nPeriod2]);
  uint32_t code = atol(itemValues[item][nCode2]);
  sKakuOld(status, period, code);
}

/** serial2 input from WS (websSocket) module **/
void readSerial2() {
  while (Serial2.available()) {
    char inC = Serial2.read();
    if (inC == '\n') {
      char _msg[strlen(inBufSer) + 1];
      strcpy(_msg, inBufSer);
      if (_msg[0] == '|' && _msg[strlen(_msg) - 1] == '|') {  // start/end with "|"?
        char ws[5] = {_msg[7], _msg[8], _msg[9], _msg[10]};  // is "|WS|" ?
        if(strcmp(ws, "|WS|") == 0) {
          /** test format and address **/
          if (_msg[7] == '|') {
            char c[7] = {_msg[1], _msg[2], _msg[3],_msg[4], _msg[5], _msg[6]};  // address test on: MAC, ID1 and ID2
            if ((strcmp(c, now_Address) == 0) || (strcmp(c, now_PairedAddr) == 0)) {
              if (now_msgInQueue == now_msgQueueLen) {  // queue full?
                for (uint8_t i = 1; i < now_msgQueueLen; i++) strcpy(now_msgQueue[i - 1], now_msgQueue[i]); // delete oldest msg
                now_msgInQueue--;
              }
              strcpy(now_msgQueue[now_msgInQueue++], _msg);
              if (strcmp(c, now_PairedAddr) == 0) sendTo(_msg);  // send to paired Nodes
            }
            else sendTo(_msg);  // send to other Nodes
          }
        }
      }
      inBufSer[0] = '\0';
    }
    else {
      if (inC < ' ' || inC > '|') inBufSer[0] = '\0';  // cut false char
      else strcat(inBufSer, String(inC).c_str());
    }
   yield();
  }
}

#endif
