////////////////////////////////////////////
// specify the funcion Items of this Node //
////////////////////////////////////////////

// ==========================================================//
//  the functions of this Node                               //
//  kitchen boiler                                           //
// ==========================================================//
//  edit the specific function Items code in:                //
//  -----------------------------------------                //
//  "myST_Start.h"   is called from setup()                  //
//  "myST_Cmd.h"     is called on a received command         //
//  "myST_Actiont.h" is, once per minute, called from loop() //
//  if needed: extra onStart/Command/Action functions        //
//    bellow in this header file                             //
// ==========================================================//

#ifndef _MY_ST_NODE_H_
#define _MY_ST_NODE_H_

// INSTAL the OneWire-master library from https://github.com/PaulStoffregen/OneWire
#include <sTempDS18B20.h>  // use only a DS18B20 and write it in the parasitic power mode
#include "myST_Gen.h"

const uint8_t OTA_Pin = 15,  // OTA upload: touch on PIN<n> or use the WS command "|OTA|" ...
              threshold = 40,  // OTA and home touch threshold
              tempPin = 19,
              relayPin = 23,
              redPin = 17,
              greenPin = 16,
              overShoot = 3, // overshoot of heating element
              maxTemp = 80;  // boiler OFF on max temp

time_t schedule[2],
       msTmp = 0;
uint8_t targetTmp = 6,         // set real value in onStart() and ...
        boilerTmp = maxTemp;  // ... used in fromLoop() and onStart(), but on a ESP(re)start boile OFF
boolean isfahrenheit = false;  // false: celsius

// a @ sign mean do NOT change this name
// =====================================
enum {  // this Node Items
  nBoiler,  // Item 1
  nItems,   //@ the number of items in use on this Node
};
enum {  // basic field names
  nName,    //@
  nMode,    //@
  nStatus,  //@
  nPause    //@
};
enum {  // Item 1 special field names
  nTmpLow1 = nPause + 1,  // off temp
  nTmpHigh1,              // on temp
  nHourOn1,               // high time start
  nHourOff1,              // high time stop
  nFields1  //@
};  // I use multiple high-low temp times, high on nHigh[n]-temp-hour and after nHiMinu (minutes) nTmpLo

const uint8_t nFields = nFields1,  // the maximum number of used Node Items fields
              itemFields[nItems] = {nFields1},  // the number of fields Items[0], Item[1}, ... , Item[n]
              nMaxChars = (10 + 1);  // max number of chars in the <...Fields[]> and <...Values[]>
const uint16_t maxItemLen = nFields * nMaxChars + nFields + 1,
               eepromAddr[nItems] = {10};  // Item EEPROM start addresses (start after: OTA[0], address[1-6], and useDST[7])
char fldIn[nFields][nMaxChars],  // parse received message to fieldnames an fieldValues 
     valIn[nFields][nMaxChars];  //   ...
const char itemNames[nItems][nFields][nMaxChars] =  {"name",   "mode", "status", "pause", "tmpLow", "tmpHigh", "hourOnn", "hourOff"},
           itemValOrg[nItems][nFields][nMaxChars] = {"boiler", "1",    "0",      "0",     "6",      "60",      "9",       "21"};
char itemValues[nItems][nFields][nMaxChars];               // mode 1: ON,   0: OFF (temp tmpOff)
char cmdList[nItems][12] = {"TEMP|TARGET"};  // set additional command list or "nothing", put code in myST_Cmd.h->onCommand() 

/** general used functions **/
void fromSetup();  // wath to do from setup()?
void fromLoop_1();   // wath to do from loop() before onAction()?
void fromLoop_2();   // wath to do from loop() after onAction()?
uint8_t parseFldVal(char * mesage);  // used by onCommand() to parse the input char string
///////////////////////////////////////
/** spesial functions for this Thing **/
void doBoiler(uint8_t on); // boiler and LED ON/OFF
uint8_t getTmp();  // from DS18B20 sensor
void doSchedule();
///////////////////////////////////////

/** to do from setup() here **/
void fromSetup() {
  /** init this Node Item(s) pins and ... here **/
  pinMode(relayPin, OUTPUT);
  pinMode(redPin, OUTPUT);
  pinMode(greenPin, OUTPUT);
  doBoiler(OFF);
  /** init DS18B20 digital temperature sensor **/
  sInitDS(tempPin);
  getTmp();
  delay(500);
  getTmp();
  doSchedule();
}
 
/** to do from loop() here **/
void fromLoop_1() {
  // before onAction()
  if (msTmp < millis()) {  // 20 sec measurement
    boilerTmp = getTmp();
    if (boilerTmp >= maxTemp || boilerTmp >= targetTmp) doBoiler(OFF);
    msTmp = millis() + 20000;
  }
}

/** to do from loop() here **/
void fromLoop_2() {
  // after onAction()
}

/** parse the received message to fieldnames an fieldValues **/ 
uint8_t parseFldVal(char * mesage) {
  uint8_t n = 0;
  char * v;
  /** get the number of fields **/
  v = strchr(mesage, '|');
  while (v != NULL) {
    n++;
    v = strchr(v + 1, '|');
  }
  if (n > 3 && (n % 2 != 0)) { // <= 3:no data, 'n' not neg. wrong format
    n = (n - 1) / 2;;
    uint8_t i = 0;
    /** get the fields **/
    v = strtok (mesage, "|");
    while (v != NULL) {
      strcpy(fldIn[i], v);
      v = strtok (NULL, "|");
      strcpy(valIn[i++], v);
      v = strtok (NULL, "|");
    }
  }
  else n = 0;
  return n;
}

///////////////////////////////////////
// == extra Node function here == //
void doBoiler(uint8_t on) {  // on: LOW = ON, !on:HIGH = OFF
  digitalWrite(relayPin, !on);
  digitalWrite(redPin, !on);
  digitalWrite(greenayPin, on);
  strcpy(itemValues[nBoiler][nStatus], String(on).c_str());
}

uint8_t getTmp() {
  double c = sGetTempDS(); 
  if (isfahrenheit) return (uint8_t)(c * 1.8 + 32.0);
  else return (uint8_t)c;
}

void doSchedule() {
  schedule[0] = atoi(itemValues[nBoiler][nHourOn1]) * 3600;   // high ...
  schedule[1] = atoi(itemValues[nBoiler][nHourOff1]) * 3600;  // ... to low dayTime in sec
}

#endif
